/************************************************************************
 *     ACPI hotkey driver foe Panasonic Let's Note Light (CF-R3)
 *
 * This driver tested Panasonic Let's Note Light CF-R3 only.
 * CF-[RTWY]2 may be works, but not tested.
 *
 * By: YOKOTA Hiroshi
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>

#include <acpi/acpi.h>
#include <acpi/acpi_drivers.h>
#include <acpi/acpi_bus.h>

MODULE_AUTHOR("YOKOTA Hiroshi");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("ACPI hotkey driver for Panasonic Let's Note Light");

#define PANA_HKEY_NAME		"Panasonic Let's Note Light"
#define PANA_HKEY_CLASS		"hotkey"
#define PANA_HKEY_DEVICE_NAME	"Hotkey"
#define PANA_HKEY_HID		"MAT0019"

/*
 * System infomeation data that held on "\_SB.HKEY.SIFR"
 */
#define SIFR_BRIGHT_AC     4     /* Back light brightness (AC) */
#define SIFR_BRIGHT_BATT   7     /* Back light brightness (Battery) */
#define SIFR_SOUND_DISABLE 8     /* Sound mute */
#define SIFR_FN_LOCK       0x80  /* Fn key shift lock mode */ // xxx

/*
 * Hotkey number
 */
/* key release */
#define HKEY_BRIGHT_DOWN   1
#define HKEY_BRIGHT_UP	   2
#define HKEY_CHANGE_GRFX   8 // XXX
#define HKEY_VOL_MUTE	   4
#define HKEY_VOL_DOWN	   5
#define HKEY_VOL_UP	   6
#define HKEY_RESUME_MEM	   7
#define HKEY_CHANGE_CPU	   3 // XXX
#define HKEY_BATT_STAT	   9
#define HKEY_RESUME_DISK  10
/* key push */
#define KEY_DOWN(key) ((key) + 0x80)
/* ------- */
#define HKEY_FN_LOCK_ON  0x30
#define HKEY_FN_LOCK_OFF 0x31

/*
 * brightness
 */
enum {
	BRIGHT_UP, BRIGHT_DOWN, BRIGHT_NONE,
};
#define BRIGHT_MAX 0x14 /* = 20 */
#define BRIGHT_MIN 0

#define PRINTK(format, ...) printk("panasonic_acpi: " format, ##__VA_ARGS__)

/*------------------------------------------------------------------------*/

static int pana_hkey_add(struct acpi_device *device);
static int pana_hkey_remove(struct acpi_device *device, int type);

static struct acpi_driver pana_hkey_driver = {
	.name =    PANA_HKEY_NAME,
	.class =   PANA_HKEY_CLASS,
	.ids =     PANA_HKEY_HID,
	.ops = {
		.add =    pana_hkey_add,
		.remove = pana_hkey_remove,
	},
};

typedef struct _pana_hotkey_data {
	struct acpi_device *device;
	acpi_handle handle;
} pana_hotkey_data;

/*
 * returns TRUE if uses in battery mode.
 */
#define BST_SIZE 8
static int pana_hkey_is_battery(void)
{
	acpi_status status = AE_OK;
	union acpi_object batt[BST_SIZE] = {
		{ .package.type     = ACPI_TYPE_PACKAGE,
		  .package.count    = ARRAY_SIZE(batt),
		  .package.elements = batt,
		},
		{},
	};
	struct acpi_buffer output = {
		.length  = sizeof(batt),
		.pointer = batt,
	};
	int batt_status;

	status = acpi_evaluate_object(NULL, "\\_SB.BATA._BST", NULL, &output);
	batt_status = batt[1].integer.value;

//	printk("b stat: %d\n", status);
	return (batt_status & 0x01) ? TRUE : FALSE;
}

/* Call ACPI "\_SB.HKEY.SSET" method to set system information */
static acpi_status pana_hkey_sset(acpi_handle handle, int function, int value)
{
	union acpi_object params[] = {
		{ .integer.type	= ACPI_TYPE_INTEGER, .integer.value = function, },
		{ .integer.type	= ACPI_TYPE_INTEGER, .integer.value = value,    },
	};
	struct acpi_object_list input = {
		.count   = ARRAY_SIZE(params),
		.pointer = params,
	};
	acpi_status status = AE_OK;

	status = acpi_evaluate_object(handle, "SSET", &input, NULL);
	return status;
}

/* Call ACPI "\_SB.HKEY.SINF" method to get saved system information */
/*------------------
 * Actual SIFR size
 * CF-R2:  9
 * CF-R3: 10
 *-----------------*/
#define SIFR_SIZE 0x10
static int pana_hkey_sinf(acpi_handle handle, int function)
{
	acpi_status status = AE_OK;
	union acpi_object  sysinf[SIFR_SIZE] = {
		{ .package.type     = ACPI_TYPE_PACKAGE,
		  .package.count    = ARRAY_SIZE(sysinf),
		  .package.elements = sysinf,
		},
		{},
	};
	struct acpi_buffer output = {
		.length  = sizeof(sysinf),
		.pointer = sysinf,
	};
//	int i;

	status = acpi_evaluate_object(handle, "SINF", NULL, &output);

/*
	printk("sifr:");
	for (i = 0 ; i < SIFR_SIZE; i++) {
		printk(" %x", (int)sysinf[i + 1].integer.value);
	}
	printk("\n");
*/

	return sysinf[function + 1].integer.value;
}

/* Call ACPI "\_SB.HKEY.HINF" method to get pushed hotkey number */
static int pana_hkey_hinf(acpi_handle handle)
{
	union acpi_object  hkey_info = {
		.type = ACPI_TYPE_INTEGER,
	};
	struct acpi_buffer output = {
		.length  = sizeof(hkey_info),
		.pointer = &hkey_info,
	};

	acpi_evaluate_object(handle, "HINF", NULL, &output);

	return hkey_info.integer.value;
}

/**************************************************************************/

static acpi_status pana_hkey_change_bright(acpi_handle handle, int key)
{
	acpi_status status = AE_OK;
	int bright;
	int bright_function;

	bright_function = pana_hkey_is_battery() ? SIFR_BRIGHT_BATT : SIFR_BRIGHT_AC;

	bright = pana_hkey_sinf(handle, bright_function);

//	printk("br: 0x%x %d\n", bright, pana_hkey_is_battery());

	switch(key) {
	case BRIGHT_UP:
		bright ++;
		if (bright > BRIGHT_MAX) { bright = BRIGHT_MAX; }
		break;

	case BRIGHT_DOWN:
		bright --;
		if (bright < BRIGHT_MIN) { bright = BRIGHT_MIN; }
		break;

	case BRIGHT_NONE:
	default:
		break;
	}

	status = pana_hkey_sset(handle, bright_function, bright);

	return status;
}

static acpi_status pana_hkey_vol_mute(acpi_handle handle)
{
	acpi_status status = AE_OK;
	int vol_mute;

	vol_mute = pana_hkey_sinf(handle, SIFR_SOUND_DISABLE);
//	printk("out: 0x%x\n", vol_mute);

	status = pana_hkey_sset(handle, SIFR_SOUND_DISABLE,
				(vol_mute == 0 ? 1 : 0));
	return status;
}

/* XXX */
/* Change VGA output. This is not legal way to change VGA. Be careful! */
static acpi_status pana_hkey_change_grfx(void)
{
	acpi_status status = AE_OK;
	union acpi_object params[] = {
		{ .integer.type = ACPI_TYPE_INTEGER, .integer.value = 0xb9, },
		{ .integer.type = ACPI_TYPE_INTEGER, .integer.value = 0x6,  },
		{ .integer.type = ACPI_TYPE_INTEGER, .integer.value = 0,    },
		{ .integer.type = ACPI_TYPE_INTEGER, .integer.value = 0,    },
		{ .integer.type = ACPI_TYPE_INTEGER, .integer.value = 0,    },
	};
	struct acpi_object_list input = {
		.count   = ARRAY_SIZE(params),
		.pointer = params,
	};

	status = acpi_evaluate_object(NULL, "\\_SB.PSRV", &input, NULL);

	return status;
}

/***************************************************************************/

/*
 * Hotkey notifier entry point
 */
static void pana_hkey_notify(acpi_handle handle, u32 event, void *driver_data)
{
	pana_hotkey_data   *data = driver_data;
	acpi_status	   status = AE_OK;
	int knum;

	if (data  == NULL) { return; }
	if (event != 0x80) { return; }
//	printk("%s\n", __func__);
//	printk("handle %p: ev 0x%x\n", handle, event);

	knum = pana_hkey_hinf(handle);

//	printk("hkey:0x%x\n", knum);

	switch (knum) {
	case KEY_DOWN(HKEY_BRIGHT_DOWN):
		status = pana_hkey_change_bright(handle, BRIGHT_DOWN);
		break;

	case KEY_DOWN(HKEY_BRIGHT_UP):
		status = pana_hkey_change_bright(handle, BRIGHT_UP);
		break;

	case KEY_DOWN(HKEY_VOL_MUTE):
		status = pana_hkey_vol_mute(handle);
		break;

	case KEY_DOWN(HKEY_CHANGE_GRFX):
		status = pana_hkey_change_grfx();
		break;

	default:
		break;
	}

//	printk("stat: %d\n", status);

	/* send event notification to userspace */
	acpi_bus_generate_event(data->device, event, knum);

	return;
}

/**************************************************************************/


static int __init pana_hkey_add(struct acpi_device *device)
{
	pana_hotkey_data *data   = NULL;
	acpi_status       status = AE_OK;

//	printk("%s device=0x%p\n", __func__, device);

	if (device == NULL) {
		return -EINVAL;
	}

	data = kmalloc(sizeof(pana_hotkey_data), GFP_KERNEL);
	if (data == NULL) {
		return -ENOMEM;
	}
	memset(data, 0, sizeof(pana_hotkey_data));

	data->handle = device->handle;
	strcpy(acpi_device_name(device),  PANA_HKEY_DEVICE_NAME);
	strcpy(acpi_device_class(device), PANA_HKEY_CLASS);
	acpi_driver_data(device) = data;
	data->device = device;

	status = acpi_install_notify_handler(data->handle, ACPI_DEVICE_NOTIFY,
					     pana_hkey_notify, data);

//	printk("%s: device=0x%p data=0x%p data->device=0x%p stat=%d\n",
//	       __func__, device, data, data->device, status);
	return 0;
}

static int pana_hkey_remove(struct acpi_device *device, int type)
{
	pana_hotkey_data *data   = NULL;
	acpi_status       status = AE_OK;

//	printk("%s\n", __func__);

	if ((device		      == NULL) ||
	    (acpi_driver_data(device) == NULL)) {
		return -EINVAL;
	}

	data = acpi_driver_data(device);

	status = acpi_remove_notify_handler(data->handle, ACPI_DEVICE_NOTIFY,
					    pana_hkey_notify);
//	printk("%s: stat=%x\n", __func__, status);
	if (ACPI_FAILURE(status)) {
		PRINTK("remove fail\n");
	}

	kfree(data);
	data = NULL;

	return 0;
}


/**************************************************************************/

static int __init pana_acpi_init(void)
{
	int result;

	PRINTK("loading...\n");

	result = acpi_bus_register_driver(&pana_hkey_driver);
	if (result < 0) {
		PRINTK("load fail\n");
		/* remove_proc_entry(PROC_PANASONIC, acpi_root_dir);*/
		return -ENODEV;
	}

	return  0;
}

static void __exit pana_acpi_exit(void)
{
	PRINTK("unloading...\n");

	acpi_bus_unregister_driver(&pana_hkey_driver);

	return;
}

module_init(pana_acpi_init);
module_exit(pana_acpi_exit);

/* end */
